#!/var/folders/nz/j6p8yfhx1mv_0grj5xl4650h0000gp/T/abs_a39nh5lowr/croot/gdal-split_1691162687573/_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pla/bin/python

import sys

from osgeo.gdal import deprecation_warn

# import osgeo_utils.gdal_polygonize as a convenience to use as a script
from osgeo_utils.gdal_polygonize import *  # noqa
from osgeo_utils.gdal_polygonize import main

deprecation_warn("gdal_polygonize")
sys.exit(main(sys.argv))



#!/usr/bin/env python3
###############################################################################
# $Id$
#
# Project:  GDAL Python samples
# Purpose:  Script to write out ASCII GRD rasters (used in Golden Software
#           Surfer)
#           from any source supported by GDAL.
# Author:   Andrey Kiselev, dron@remotesensing.org
#
###############################################################################
# Copyright (c) 2003, Andrey Kiselev <dron@remotesensing.org>
# Copyright (c) 2009, Even Rouault <even dot rouault at spatialys.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
###############################################################################

import sys

from osgeo import gdal

gdal.TermProgress = gdal.TermProgress_nocb

# =============================================================================


def Usage():
    print("Usage: gdal2grd.py [-b band] [-quiet] infile outfile")
    print("Write out ASCII GRD rasters (used in Golden Software Surfer)")
    print("")
    print("  -b band	    Select a band number to convert (1 based)")
    print("  -quiet	    Do not report any diagnostic information")
    print("  infile	    Name of the input GDAL supported file")
    print("  outfile	    Name of the output GRD file")
    print("")
    return 2


def main(argv=sys.argv):
    infile = None
    outfile = None
    iBand = 1
    quiet = 0

    # Parse command line arguments.
    i = 1
    while i < len(sys.argv):
        arg = sys.argv[i]

        if arg == "-b":
            i = i + 1
            iBand = int(sys.argv[i])

        elif arg == "-quiet":
            quiet = 1

        elif infile is None:
            infile = arg

        elif outfile is None:
            outfile = arg

        else:
            Usage()

        i = i + 1

    if infile is None:
        return Usage()
    if outfile is None:
        return Usage()

    indataset = gdal.Open(infile, gdal.GA_ReadOnly)
    if infile is None:
        print("Cannot open", infile)
        return 2
    geotransform = indataset.GetGeoTransform()
    band = indataset.GetRasterBand(iBand)
    if band is None:
        print("Cannot load band", iBand, "from the", infile)
        return 2

    if not quiet:
        print(
            "Size is ",
            indataset.RasterXSize,
            "x",
            indataset.RasterYSize,
            "x",
            indataset.RasterCount,
        )
        print("Projection is ", indataset.GetProjection())
        print("Origin = (", geotransform[0], ",", geotransform[3], ")")
        print("Pixel Size = (", geotransform[1], ",", geotransform[5], ")")
        print(
            "Converting band number",
            iBand,
            "with type",
            gdal.GetDataTypeName(band.DataType),
        )

    # Header printing
    fpout = open(outfile, "wt")
    fpout.write("DSAA\n")
    fpout.write(str(band.XSize) + " " + str(band.YSize) + "\n")
    fpout.write(
        str(geotransform[0] + geotransform[1] / 2)
        + " "
        + str(geotransform[0] + geotransform[1] * (band.XSize - 0.5))
        + "\n"
    )
    if geotransform[5] < 0:
        fpout.write(
            str(geotransform[3] + geotransform[5] * (band.YSize - 0.5))
            + " "
            + str(geotransform[3] + geotransform[5] / 2)
            + "\n"
        )
    else:
        fpout.write(
            str(geotransform[3] + geotransform[5] / 2)
            + " "
            + str(geotransform[3] + geotransform[5] * (band.YSize - 0.5))
            + "\n"
        )
    fpout.write(
        str(band.ComputeRasterMinMax(0)[0])
        + " "
        + str(band.ComputeRasterMinMax(0)[1])
        + "\n"
    )

    for i in range(band.YSize - 1, -1, -1):
        scanline = band.ReadAsArray(0, i, band.XSize, 1, band.XSize, 1)
        j = 0
        while j < band.XSize:
            fpout.write(str(scanline[0, j]))
            j = j + 1
            if j % 10:  # Print no more than 10 values per line
                fpout.write(" ")
            else:
                fpout.write("\n")
        fpout.write("\n")

        # Display progress report on terminal
        if not quiet:
            gdal.TermProgress(float(band.YSize - i) / band.YSize)
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# ******************************************************************************
#  $Id$
#
#  Project:  GDAL Python Interface
#  Purpose:  Application for applying sieve filter to raster data.
#  Author:   Frank Warmerdam, warmerdam@pobox.com
#
# ******************************************************************************
#  Copyright (c) 2008, Frank Warmerdam
#  Copyright (c) 2009-2010, Even Rouault <even dot rouault at spatialys.com>
#  Copyright (c) 2021, Idan Miara <idan@miara.com>
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
# ******************************************************************************

import sys
from typing import Optional

from osgeo import gdal
from osgeo_utils.auxiliary.base import PathLikeOrStr
from osgeo_utils.auxiliary.util import GetOutputDriverFor


def Usage():
    print(
        """Usage: gdal_sieve [-q] [-st threshold] [-4] [-8] [-o name=value]
           srcfile [-nomask] [-mask filename] [-of format] [dstfile]"""
    )
    return 2


def main(argv=sys.argv):
    threshold = 2
    connectedness = 4
    quiet = False
    src_filename = None

    dst_filename = None
    driver_name = None

    mask = "default"

    argv = gdal.GeneralCmdLineProcessor(argv)
    if argv is None:
        return 0

    # Parse command line arguments.
    i = 1
    while i < len(argv):
        arg = argv[i]

        if arg == "-of" or arg == "-f":
            i = i + 1
            driver_name = argv[i]

        elif arg == "-4":
            connectedness = 4

        elif arg == "-8":
            connectedness = 8

        elif arg == "-q" or arg == "-quiet":
            quiet = True

        elif arg == "-st":
            i = i + 1
            threshold = int(argv[i])

        elif arg == "-nomask":
            mask = "none"

        elif arg == "-mask":
            i = i + 1
            mask = argv[i]

        elif arg == "-mask":
            i = i + 1
            mask = argv[i]

        elif arg[:2] == "-h":
            return Usage()

        elif src_filename is None:
            src_filename = argv[i]

        elif dst_filename is None:
            dst_filename = argv[i]

        else:
            return Usage()

        i = i + 1

    if src_filename is None:
        return Usage()

    return gdal_sieve(
        src_filename=src_filename,
        dst_filename=dst_filename,
        driver_name=driver_name,
        mask=mask,
        threshold=threshold,
        connectedness=connectedness,
        quiet=quiet,
    )


def gdal_sieve(
    src_filename: Optional[str] = None,
    dst_filename: PathLikeOrStr = None,
    driver_name: Optional[str] = None,
    mask: str = "default",
    threshold: int = 2,
    connectedness: int = 4,
    quiet: bool = False,
):
    # =============================================================================
    # 	Verify we have next gen bindings with the sievefilter method.
    # =============================================================================
    try:
        gdal.SieveFilter
    except AttributeError:
        print("")
        print('gdal.SieveFilter() not available.  You are likely using "old gen"')
        print("bindings or an older version of the next gen bindings.")
        print("")
        return 1

    # =============================================================================
    # Open source file
    # =============================================================================

    if dst_filename is None:
        src_ds = gdal.Open(src_filename, gdal.GA_Update)
    else:
        src_ds = gdal.Open(src_filename, gdal.GA_ReadOnly)

    if src_ds is None:
        print("Unable to open %s " % src_filename)
        return 1

    srcband = src_ds.GetRasterBand(1)

    if mask == "default":
        maskband = srcband.GetMaskBand()
    elif mask == "none":
        maskband = None
    else:
        mask_ds = gdal.Open(mask)
        maskband = mask_ds.GetRasterBand(1)

    # =============================================================================
    #       Create output file if one is specified.
    # =============================================================================

    if dst_filename is not None:
        if driver_name is None:
            driver_name = GetOutputDriverFor(dst_filename)

        drv = gdal.GetDriverByName(driver_name)
        dst_ds = drv.Create(
            dst_filename, src_ds.RasterXSize, src_ds.RasterYSize, 1, srcband.DataType
        )
        wkt = src_ds.GetProjection()
        if wkt != "":
            dst_ds.SetProjection(wkt)
        gt = src_ds.GetGeoTransform(can_return_null=True)
        if gt is not None:
            dst_ds.SetGeoTransform(gt)

        dstband = dst_ds.GetRasterBand(1)
        nodata = srcband.GetNoDataValue()
        if nodata is not None:
            dstband.SetNoDataValue(nodata)
    else:
        dstband = srcband

    # =============================================================================
    # Invoke algorithm.
    # =============================================================================

    if quiet:
        prog_func = None
    else:
        prog_func = gdal.TermProgress_nocb

    result = gdal.SieveFilter(
        srcband, maskband, dstband, threshold, connectedness, callback=prog_func
    )

    src_ds = None
    dst_ds = None
    mask_ds = None

    return result


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ******************************************************************************
#  $Id$
#
#  Project:  CFS OGC MapServer
#  Purpose:  Script to create WKT and PROJ.4 dictionaries for EPSG GCS/PCS
#            codes.
#  Author:   Frank Warmerdam, warmerdam@pobox.com
#
# ******************************************************************************
#  Copyright (c) 2001, Frank Warmerdam
#  Copyright (c) 2009-2010, 2019, Even Rouault <even dot rouault at spatialys.com>
#  Copyright (c) 2021, Idan Miara <idan@miara.com>
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
# ******************************************************************************

import sys
from typing import Optional

from osgeo import gdal, osr
from osgeo_utils.auxiliary.gdal_argparse import GDALArgumentParser, GDALScript


def Usage():
    print(
        f"Usage: {sys.argv[0]} -- This is a sample. Read source to know how to use. --"
    )
    return 2


def trHandleCode(set_srid, srs, auth_name, code, deprecated, output_format):
    if output_format == "-pretty_wkt":
        print("%s:%s" % (auth_name, str(code)))
        print(srs.ExportToPrettyWkt())

    if output_format == "-xml":
        print(srs.ExportToXML())

    if output_format == "-wkt":
        print(f"EPSG:{code}")
        print(srs.ExportToWkt())

    if output_format == "-proj4":
        out_string = srs.ExportToProj4()

        name = srs.GetName()

        print("# %s" % name)
        if out_string.find("+proj=") > -1:
            print("<%s> %s <>" % (str(code), out_string))
        else:
            print(
                "# Unable to translate coordinate system "
                "%s:%s into PROJ.4 format." % (auth_name, str(code))
            )
            print("#")

    if output_format == "-postgis":

        if code in set_srid:
            if auth_name == "ESRI":
                if int(code) < 32767:
                    return
        assert code not in set_srid, (auth_name, code)
        set_srid.add(code)

        name = srs.GetName()
        if deprecated and "deprecated" not in name:
            name += " (deprecated)"
        wkt = srs.ExportToWkt()
        proj4text = srs.ExportToProj4()

        print("---")
        print("--- %s %s : %s" % (auth_name, str(code), name))
        print("---")

        if proj4text is None or len(proj4text) == 0:
            print("-- (unable to translate to PROJ.4)")
        else:
            wkt = gdal.EscapeString(wkt, scheme=gdal.CPLES_SQL)
            proj4text = gdal.EscapeString(proj4text, scheme=gdal.CPLES_SQL)
            print(
                'INSERT INTO "spatial_ref_sys" ("srid","auth_name","auth_srid","srtext","proj4text") VALUES (%d,\'%s\',%d,\'%s\',\'%s\');'
                % (int(code), auth_name, int(code), wkt, proj4text)
            )

    # INGRES COPY command input.
    if output_format == "-copy":

        try:
            wkt = srs.ExportToWkt()
            proj4text = srs.ExportToProj4()

            print(
                "%s\t%d%s\t%s\t%d%s\t%d%s\n"
                % (
                    str(code),
                    4,
                    auth_name,
                    str(code),
                    len(wkt),
                    wkt,
                    len(proj4text),
                    proj4text,
                )
            )
        except Exception:
            pass


def epsg_tr(output_format: str = "-pretty_wkt", authority: Optional[str] = None):
    # Output BEGIN transaction for PostGIS
    if output_format == "-postgis":
        print("BEGIN;")

    # loop over all codes to generate output

    if authority:
        authorities = [authority]
    elif output_format == "-postgis":
        authorities = ["EPSG", "ESRI"]
    else:
        authorities = ["EPSG", "ESRI", "IGNF"]

    set_srid = set()
    for authority in authorities:
        if authority in ("EPSG", "ESRI"):
            set_codes_geographic = set()
            set_codes_geographic_3d = set()
            set_codes_projected = set()
            set_codes_geocentric = set()
            set_codes_compound = set()
            set_deprecated = set()

            for crs_info in osr.GetCRSInfoListFromDatabase(authority):
                code = int(crs_info.code)
                if crs_info.type == osr.OSR_CRS_TYPE_COMPOUND:
                    set_codes_compound.add(code)
                elif crs_info.type == osr.OSR_CRS_TYPE_GEOGRAPHIC_3D:
                    set_codes_geographic_3d.add(code)
                elif crs_info.type == osr.OSR_CRS_TYPE_GEOGRAPHIC_2D:
                    set_codes_geographic.add(code)
                elif crs_info.type == osr.OSR_CRS_TYPE_PROJECTED:
                    set_codes_projected.add(code)
                elif crs_info.type == osr.OSR_CRS_TYPE_GEOCENTRIC:
                    set_codes_geocentric.add(code)

                if crs_info.deprecated:
                    set_deprecated.add(code)

            set_codes_geographic = sorted(set_codes_geographic)
            set_codes_geographic_3d = sorted(set_codes_geographic_3d)
            set_codes_projected = sorted(set_codes_projected)
            set_codes_geocentric = sorted(set_codes_geocentric)
            set_codes_compound = sorted(set_codes_compound)
            for typestr, set_codes in (
                ("Geographic 2D CRS", set_codes_geographic),
                ("Projected CRS", set_codes_projected),
                ("Geocentric CRS", set_codes_geocentric),
                ("Compound CRS", set_codes_compound),
                ("Geographic 3D CRS", set_codes_geographic_3d),
            ):
                if set_codes and output_format == "-postgis":
                    print("-" * 80)
                    print("--- " + authority + " " + typestr)
                    print("-" * 80)

                for code in set_codes:
                    srs = osr.SpatialReference()
                    srs.SetFromUserInput(authority + ":" + str(code))
                    deprecated = False
                    if code in set_deprecated:
                        deprecated = True
                    trHandleCode(
                        set_srid, srs, authority, str(code), deprecated, output_format
                    )

        else:
            for crs_info in osr.GetCRSInfoListFromDatabase(authority):
                srs = osr.SpatialReference()
                srs.SetFromUserInput(authority + ":" + crs_info.code)
                trHandleCode(
                    set_srid,
                    srs,
                    authority,
                    crs_info.code,
                    crs_info.deprecated,
                    output_format,
                )

    # Output COMMIT transaction for PostGIS
    if output_format == "-postgis":
        print("COMMIT;")
        print("VACUUM ANALYZE spatial_ref_sys;")
    return 0


class EPSG_Table(GDALScript):
    def __init__(self):
        super().__init__()
        self.title = "Create WKT and PROJ.4 dictionaries for EPSG GCS/PCS codes"

    def get_parser(self, argv) -> GDALArgumentParser:
        parser = self.parser

        parser.add_argument(
            "-authority",
            dest="authority",
            metavar="name",
            type=str,
            help="Authority name",
        )

        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "-wkt",
            const="wkt",
            dest="output_format",
            action="store_const",
            help="Well Known Text",
        )
        group.add_argument(
            "-pretty_wkt",
            const="pretty_wkt",
            dest="output_format",
            action="store_const",
            help="Pretty Well Known Text",
        )
        group.add_argument(
            "-proj4",
            const="proj4",
            dest="output_format",
            action="store_const",
            help="proj string",
        )
        group.add_argument(
            "-postgis",
            const="postgis",
            dest="output_format",
            action="store_const",
            help="postgis",
        )
        group.add_argument(
            "-xml", const="xml", dest="output_format", action="store_const", help="XML"
        )
        group.add_argument(
            "-copy",
            const="copy",
            dest="output_format",
            action="store_const",
            help="Table",
        )

        return parser

    def augment_kwargs(self, kwargs) -> dict:
        if not kwargs.get("output_format"):
            kwargs["output_format"] = "pretty_wkt"
        kwargs["output_format"] = "-" + kwargs["output_format"]
        return kwargs

    def doit(self, **kwargs):
        return epsg_tr(**kwargs)


def main(argv=sys.argv):
    if len(sys.argv) < 2:
        return Usage()
    return EPSG_Table().main(argv)


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# ******************************************************************************
#  $Id$
#
#  Project:  GDAL Python Interface
#  Purpose:  Application for filling nodata areas in a raster by interpolation
#  Author:   Frank Warmerdam, warmerdam@pobox.com
#
# ******************************************************************************
#  Copyright (c) 2008, Frank Warmerdam
#  Copyright (c) 2009-2011, Even Rouault <even dot rouault at spatialys.com>
#  Copyright (c) 2021, Idan Miara <idan@miara.com>
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
# ******************************************************************************

import sys
import textwrap
from numbers import Real
from typing import Optional

from osgeo import gdal
from osgeo_utils.auxiliary.gdal_argparse import GDALArgumentParser, GDALScript


def CopyBand(srcband, dstband):
    for line in range(srcband.YSize):
        line_data = srcband.ReadRaster(0, line, srcband.XSize, 1)
        dstband.WriteRaster(
            0, line, srcband.XSize, 1, line_data, buf_type=srcband.DataType
        )


def gdal_fillnodata(
    src_filename: Optional[str] = None,
    band_number: int = 1,
    dst_filename: Optional[str] = None,
    driver_name: str = "GTiff",
    creation_options: Optional[list] = None,
    quiet: bool = False,
    mask: str = "default",
    max_distance: Real = 100,
    smoothing_iterations: int = 0,
    options: Optional[list] = None,
):
    options = options or []
    creation_options = creation_options or []

    # =============================================================================
    # 	Verify we have next gen bindings with the sievefilter method.
    # =============================================================================
    try:
        gdal.FillNodata
    except AttributeError:
        print("")
        print('gdal.FillNodata() not available.  You are likely using "old gen"')
        print("bindings or an older version of the next gen bindings.")
        print("")
        return 1

    # =============================================================================
    # Open source file
    # =============================================================================

    if dst_filename is None:
        src_ds = gdal.Open(src_filename, gdal.GA_Update)
    else:
        src_ds = gdal.Open(src_filename, gdal.GA_ReadOnly)

    if src_ds is None:
        print("Unable to open %s" % src_filename)
        return 1

    srcband = src_ds.GetRasterBand(band_number)

    # =============================================================================
    #       Create output file if one is specified.
    # =============================================================================

    if dst_filename is not None:

        drv = gdal.GetDriverByName(driver_name)
        dst_ds = drv.Create(
            dst_filename,
            src_ds.RasterXSize,
            src_ds.RasterYSize,
            1,
            srcband.DataType,
            creation_options,
        )
        wkt = src_ds.GetProjection()
        if wkt != "":
            dst_ds.SetProjection(wkt)
        gt = src_ds.GetGeoTransform(can_return_null=True)
        if gt:
            dst_ds.SetGeoTransform(gt)

        dstband = dst_ds.GetRasterBand(1)
        ndv = srcband.GetNoDataValue()
        if ndv is not None:
            dstband.SetNoDataValue(ndv)

        color_interp = srcband.GetColorInterpretation()
        dstband.SetColorInterpretation(color_interp)
        if color_interp == gdal.GCI_PaletteIndex:
            color_table = srcband.GetColorTable()
            dstband.SetColorTable(color_table)

        CopyBand(srcband, dstband)

    else:
        dstband = srcband

    # =============================================================================
    # Invoke algorithm.
    # =============================================================================

    if quiet:
        prog_func = None
    else:
        prog_func = gdal.TermProgress_nocb

    if mask == "default":
        maskband = dstband.GetMaskBand()
    else:
        mask_ds = gdal.Open(mask)
        maskband = mask_ds.GetRasterBand(1)

    result = gdal.FillNodata(
        dstband,
        maskband,
        max_distance,
        smoothing_iterations,
        options,
        callback=prog_func,
    )

    src_ds = None
    dst_ds = None
    mask_ds = None

    return result


class GDALFillNoData(GDALScript):
    def __init__(self):
        super().__init__()
        self.title = "Fill raster regions by interpolation from edges"
        self.description = textwrap.dedent(
            """\
            It Fills selection regions (usually nodata areas)
            by interpolating from valid pixels around the edges of the area.
            Additional details on the algorithm are available in the GDALFillNodata() docs."""
        )

    def get_parser(self, argv) -> GDALArgumentParser:
        parser = self.parser

        parser.add_argument(
            "-q",
            "-quiet",
            dest="quiet",
            action="store_true",
            help="The script runs in quiet mode. "
            "The progress monitor is suppressed and routine messages are not displayed.",
        )

        parser.add_argument(
            "-md",
            dest="max_distance",
            type=float,
            default=100,
            metavar="max_distance",
            help="The maximum distance (in pixels) "
            "that the algorithm will search out for values to interpolate. "
            "The default is 100 pixels.",
        )

        parser.add_argument(
            "-si",
            dest="smoothing_iterations",
            type=int,
            default=0,
            metavar="smoothing_iterations",
            help="The number of 3x3 average filter smoothing iterations to run after the interpolation "
            "to dampen artifacts. The default is zero smoothing iterations.",
        )

        parser.add_argument(
            "-o",
            dest="options",
            type=str,
            action="extend",
            nargs="*",
            metavar="name=value",
            help="Specify a special argument to the algorithm.",
        )

        parser.add_argument(
            "-mask",
            dest="mask",
            type=str,
            metavar="filename",
            default="default",
            help="Use the first band of the specified file as a validity mask "
            "(zero is invalid, non-zero is valid).",
        )

        parser.add_argument(
            "-b",
            "-band",
            dest="band_number",
            metavar="band",
            type=int,
            default=1,
            help="The band to operate on, defaults to 1.",
        )

        parser.add_argument(
            "-of",
            dest="driver_name",
            default="GTiff",
            metavar="gdal_format",
            help="Select the output format. Use the short format name.",
        )

        parser.add_argument(
            "-co",
            dest="creation_options",
            type=str,
            action="extend",
            nargs="*",
            metavar="name=value",
            help="Creation options for the destination dataset.",
        )

        parser.add_argument(
            "src_filename",
            type=str,
            help="The source raster file used to identify target pixels. "
            "Only one band is used.",
        )

        parser.add_argument(
            "dst_filename",
            type=str,
            help="The new file to create with the interpolated result. "
            "If not provided, the source band is updated in place.",
        )

        return parser

    def doit(self, **kwargs):
        return gdal_fillnodata(**kwargs)


def main(argv=sys.argv):
    return GDALFillNoData().main(argv)


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
###############################################################################
#  $Id$
#
# Purpose:  Module for retiling (merging) tiles and building tiled pyramids
# Author:   Christian Meuller, christian.mueller@nvoe.at
# UseDirForEachRow support by Chris Giesey & Elijah Robison
#
###############################################################################
# Copyright (c) 2007, Christian Mueller
# Copyright (c) 2009-2012, Even Rouault <even dot rouault at spatialys.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
###############################################################################
from __future__ import print_function

import os
import shutil
import sys

from osgeo import gdal, ogr, osr

progress = gdal.TermProgress_nocb


class AffineTransformDecorator(object):
    """A class providing some useful methods for affine Transformations"""

    def __init__(self, transform):
        self.lrx = None
        self.lry = None
        self.geotransform = transform
        self.scaleX = self.geotransform[1]
        self.scaleY = self.geotransform[5]
        if self.scaleY > 0:
            self.scaleY *= -1
        self.ulx = self.geotransform[0]
        self.uly = self.geotransform[3]

    def pointsFor(self, width, height):
        xlist = []
        ylist = []
        w = self.scaleX * width
        h = self.scaleY * height

        xlist.append(self.ulx)
        ylist.append(self.uly)
        xlist.append(self.ulx + w)
        ylist.append(self.uly)
        xlist.append(self.ulx + w)
        ylist.append(self.uly + h)
        xlist.append(self.ulx)
        ylist.append(self.uly + h)
        return [xlist, ylist]


class DataSetCache(object):
    """A class for caching source tiles"""

    def __init__(self):
        self.cacheSize = 8
        self.queue = []
        self.dict = {}

    def get(self, name):

        if name in self.dict:
            return self.dict[name]
        result = gdal.Open(name)
        if result is None:
            print("Error opening: %s" % NameError)
            return 1
        if len(self.queue) == self.cacheSize:
            toRemove = self.queue.pop(0)
            del self.dict[toRemove]
        self.queue.append(name)
        self.dict[name] = result
        return result

    def __del__(self):
        for dataset in self.dict.values():
            del dataset
        del self.queue
        del self.dict


class tile_info(object):
    """A class holding info how to tile"""

    def __init__(self, xsize, ysize, tileWidth, tileHeight, overlap):
        self.width = xsize
        self.height = ysize
        self.tileWidth = tileWidth
        self.tileHeight = tileHeight
        self.countTilesX = 1
        if xsize > tileWidth:
            self.countTilesX += int(
                (xsize - tileWidth + (tileWidth - overlap) - 1) / (tileWidth - overlap)
            )
        self.countTilesY = 1
        if ysize > tileHeight:
            self.countTilesY += int(
                (ysize - tileHeight + (tileHeight - overlap) - 1)
                / (tileHeight - overlap)
            )
        self.overlap = overlap

    def report(self):
        print("tileWidth:   %d" % self.tileWidth)
        print("tileHeight:  %d" % self.tileHeight)
        print("countTilesX: %d" % self.countTilesX)
        print("countTilesY: %d" % self.countTilesY)
        print("overlap:     %d" % self.overlap)


class mosaic_info(object):
    """A class holding information about a GDAL file or a GDAL fileset"""

    def __init__(self, filename, inputDS):
        """
        Initialize mosaic_info from filename

        filename -- Name of file to read.

        """
        self.TempDriver = gdal.GetDriverByName("MEM")
        self.filename = filename
        self.cache = DataSetCache()
        self.ogrTileIndexDS = inputDS

        self.ogrTileIndexDS.GetLayer().ResetReading()
        feature = self.ogrTileIndexDS.GetLayer().GetNextFeature()
        imgLocation = feature.GetField(0)

        fhInputTile = self.cache.get(imgLocation)

        self.bands = fhInputTile.RasterCount
        self.band_type = fhInputTile.GetRasterBand(1).DataType
        self.projection = fhInputTile.GetProjection()
        self.nodata = fhInputTile.GetRasterBand(1).GetNoDataValue()

        dec = AffineTransformDecorator(fhInputTile.GetGeoTransform())
        self.scaleX = dec.scaleX
        self.scaleY = dec.scaleY
        ct = fhInputTile.GetRasterBand(1).GetRasterColorTable()
        if ct is not None:
            self.ct = ct.Clone()
        else:
            self.ct = None
        self.ci = [0] * self.bands
        for iband in range(self.bands):
            self.ci[iband] = fhInputTile.GetRasterBand(
                iband + 1
            ).GetRasterColorInterpretation()

        extent = self.ogrTileIndexDS.GetLayer().GetExtent()
        self.ulx = extent[0]
        self.uly = extent[3]
        self.lrx = extent[1]
        self.lry = extent[2]

        self.xsize = int(round((self.lrx - self.ulx) / self.scaleX))
        self.ysize = abs(int(round((self.uly - self.lry) / self.scaleY)))

    def __del__(self):
        del self.cache
        del self.ogrTileIndexDS

    def getDataSet(self, minx, miny, maxx, maxy):

        self.ogrTileIndexDS.GetLayer().ResetReading()
        self.ogrTileIndexDS.GetLayer().SetSpatialFilterRect(minx, miny, maxx, maxy)
        features = []
        envelope = None
        while True:
            feature = self.ogrTileIndexDS.GetLayer().GetNextFeature()
            if feature is None:
                break
            features.append(feature)
            if envelope is None:
                envelope = feature.GetGeometryRef().GetEnvelope()
            else:
                featureEnv = feature.GetGeometryRef().GetEnvelope()
                envelope = (
                    min(featureEnv[0], envelope[0]),
                    max(featureEnv[1], envelope[1]),
                    min(featureEnv[2], envelope[2]),
                    max(featureEnv[3], envelope[3]),
                )

        if envelope is None:
            return None

        # enlarge to query rect if necessary
        envelope = (
            min(minx, envelope[0]),
            max(maxx, envelope[1]),
            min(miny, envelope[2]),
            max(maxy, envelope[3]),
        )

        self.ogrTileIndexDS.GetLayer().SetSpatialFilter(None)

        # merge tiles

        resultSizeX = int((maxx - minx) / self.scaleX + 0.5)
        resultSizeY = int((miny - maxy) / self.scaleY + 0.5)

        resultDS = self.TempDriver.Create(
            "TEMP", resultSizeX, resultSizeY, self.bands, self.band_type, []
        )
        resultDS.SetGeoTransform([minx, self.scaleX, 0, maxy, 0, self.scaleY])

        for bandNr in range(1, self.bands + 1):
            t_band = resultDS.GetRasterBand(bandNr)
            if self.nodata is not None:
                t_band.Fill(self.nodata)
                t_band.SetNoDataValue(self.nodata)

        for feature in features:
            featureName = feature.GetField(0)
            sourceDS = self.cache.get(featureName)
            dec = AffineTransformDecorator(sourceDS.GetGeoTransform())

            dec.lrx = dec.ulx + sourceDS.RasterXSize * dec.scaleX
            dec.lry = dec.uly + sourceDS.RasterYSize * dec.scaleY

            # Find the intersection region
            tgw_ulx = max(dec.ulx, minx)
            tgw_lrx = min(dec.lrx, maxx)
            if self.scaleY < 0:
                tgw_uly = min(dec.uly, maxy)
                tgw_lry = max(dec.lry, miny)
            else:
                tgw_uly = max(dec.uly, maxy)
                tgw_lry = min(dec.lry, miny)

            # Compute source window in pixel coordinates.
            sw_xoff = int((tgw_ulx - dec.ulx) / dec.scaleX + 0.5)
            sw_yoff = int((tgw_uly - dec.uly) / dec.scaleY + 0.5)
            sw_xsize = (
                min(sourceDS.RasterXSize, int((tgw_lrx - dec.ulx) / dec.scaleX + 0.5))
                - sw_xoff
            )
            sw_ysize = (
                min(sourceDS.RasterYSize, int((tgw_lry - dec.uly) / dec.scaleY + 0.5))
                - sw_yoff
            )
            if sw_xsize <= 0 or sw_ysize <= 0:
                continue

            # Compute target window in pixel coordinates
            tw_xoff = int((tgw_ulx - minx) / self.scaleX + 0.5)
            tw_yoff = int((tgw_uly - maxy) / self.scaleY + 0.5)
            tw_xsize = (
                min(resultDS.RasterXSize, int((tgw_lrx - minx) / self.scaleX + 0.5))
                - tw_xoff
            )
            tw_ysize = (
                min(resultDS.RasterYSize, int((tgw_lry - maxy) / self.scaleY + 0.5))
                - tw_yoff
            )
            if tw_xsize <= 0 or tw_ysize <= 0:
                continue

            assert tw_xoff >= 0
            assert tw_yoff >= 0
            assert sw_xoff >= 0
            assert sw_yoff >= 0

            for bandNr in range(1, self.bands + 1):
                s_band = sourceDS.GetRasterBand(bandNr)
                t_band = resultDS.GetRasterBand(bandNr)
                if self.ct is not None:
                    t_band.SetRasterColorTable(self.ct)
                t_band.SetRasterColorInterpretation(self.ci[bandNr - 1])

                data = s_band.ReadRaster(
                    sw_xoff,
                    sw_yoff,
                    sw_xsize,
                    sw_ysize,
                    tw_xsize,
                    tw_ysize,
                    self.band_type,
                )
                if data is None:
                    print(gdal.GetLastErrorMsg())

                t_band.WriteRaster(tw_xoff, tw_yoff, tw_xsize, tw_ysize, data)

        return resultDS

    def closeDataSet(self, memDS):
        del memDS
        # self.TempDriver.Delete("TEMP")

    def report(self):
        print("Filename: " + self.filename)
        print("File Size: %dx%dx%d" % (self.xsize, self.ysize, self.bands))
        print("Pixel Size: %f x %f" % (self.scaleX, self.scaleY))
        print("UL:(%f,%f)   LR:(%f,%f)" % (self.ulx, self.uly, self.lrx, self.lry))


def getTileIndexFromFiles(g):
    if g.Verbose:
        print("Building internal Index for %d tile(s) ..." % len(g.Names), end=" ")

    ogrTileIndexDS = createTileIndex(
        g.Verbose, "TileIndex", g.TileIndexFieldName, None, g.TileIndexDriverTyp
    )
    for inputTile in g.Names:

        fhInputTile = gdal.Open(inputTile)
        if fhInputTile is None:
            return None

        dec = AffineTransformDecorator(fhInputTile.GetGeoTransform())
        points = dec.pointsFor(fhInputTile.RasterXSize, fhInputTile.RasterYSize)

        addFeature(
            g.TileIndexFieldName, ogrTileIndexDS, inputTile, points[0], points[1]
        )
        del fhInputTile

    if g.Verbose:
        print("finished")
    # ogrTileIndexDS.GetLayer().SyncToDisk()
    return ogrTileIndexDS


def getTargetDir(g, level=-1):
    if level == -1:
        return g.TargetDir
    return g.TargetDir + str(level) + os.sep


def tileImage(g, minfo, ti):
    """

    Tile image in mosaicinfo minfo  based on tileinfo ti

    returns list of created tiles

    """

    g.LastRowIndx = -1
    OGRDS = createTileIndex(
        g.Verbose,
        "TileResult_0",
        g.TileIndexFieldName,
        g.Source_SRS,
        g.TileIndexDriverTyp,
    )

    yRange = list(range(1, ti.countTilesY + 1))
    xRange = list(range(1, ti.countTilesX + 1))

    if not g.Quiet and not g.Verbose:
        progress(0.0)
        processed = 0
        total = len(xRange) * len(yRange)

    for yIndex in yRange:
        for xIndex in xRange:
            offsetY = (yIndex - 1) * (ti.tileHeight - ti.overlap)
            offsetX = (xIndex - 1) * (ti.tileWidth - ti.overlap)
            height = ti.tileHeight
            width = ti.tileWidth
            if g.UseDirForEachRow:
                tilename = getTileName(g, minfo, ti, xIndex, yIndex, 0)
            else:
                tilename = getTileName(g, minfo, ti, xIndex, yIndex)

            if offsetX + width > ti.width:
                width = ti.width - offsetX
            if offsetY + height > ti.height:
                height = ti.height - offsetY

            feature_only = g.Resume and os.path.exists(tilename)
            createTile(
                g, minfo, offsetX, offsetY, width, height, tilename, OGRDS, feature_only
            )

            if not g.Quiet and not g.Verbose:
                processed += 1
                progress(processed / float(total))

    if g.TileIndexName is not None:
        if g.UseDirForEachRow and not g.PyramidOnly:
            shapeName = getTargetDir(g, 0) + g.TileIndexName
        else:
            shapeName = getTargetDir(g) + g.TileIndexName
        copyTileIndexToDisk(g, OGRDS, shapeName)

    if g.CsvFileName is not None:
        if g.UseDirForEachRow and not g.PyramidOnly:
            csvName = getTargetDir(g, 0) + g.CsvFileName
        else:
            csvName = getTargetDir(g) + g.CsvFileName
        copyTileIndexToCSV(g, OGRDS, csvName)

    return OGRDS


def copyTileIndexToDisk(g, OGRDS, fileName):
    SHAPEDS = createTileIndex(
        g.Verbose,
        fileName,
        g.TileIndexFieldName,
        OGRDS.GetLayer().GetSpatialRef(),
        "ESRI Shapefile",
    )
    OGRDS.GetLayer().ResetReading()
    while True:
        feature = OGRDS.GetLayer().GetNextFeature()
        if feature is None:
            break
        newFeature = feature.Clone()
        basename = os.path.basename(feature.GetField(0))
        if g.UseDirForEachRow:
            t = os.path.split(os.path.dirname(feature.GetField(0)))
            basename = t[1] + "/" + basename
        newFeature.SetField(0, basename)
        SHAPEDS.GetLayer().CreateFeature(newFeature)
    closeTileIndex(SHAPEDS)


def copyTileIndexToCSV(g, OGRDS, fileName):
    csvfile = open(fileName, "w")
    OGRDS.GetLayer().ResetReading()
    while True:
        feature = OGRDS.GetLayer().GetNextFeature()
        if feature is None:
            break
        basename = os.path.basename(feature.GetField(0))
        if g.UseDirForEachRow:
            t = os.path.split(os.path.dirname(feature.GetField(0)))
            basename = t[1] + "/" + basename
        csvfile.write(basename)
        geom = feature.GetGeometryRef()
        coords = geom.GetEnvelope()

        for coord in coords:
            csvfile.write(g.CsvDelimiter)
            csvfile.write("%f" % coord)
        csvfile.write("\n")

    csvfile.close()


def createPyramidTile(
    g, levelMosaicInfo, offsetX, offsetY, width, height, tileName, OGRDS, feature_only
):

    temp_tilename = tileName + ".tmp"

    sx = levelMosaicInfo.scaleX * 2
    sy = levelMosaicInfo.scaleY * 2

    dec = AffineTransformDecorator(
        [
            levelMosaicInfo.ulx + offsetX * sx,
            sx,
            0,
            levelMosaicInfo.uly + offsetY * sy,
            0,
            sy,
        ]
    )

    if OGRDS is not None:
        points = dec.pointsFor(width, height)
        addFeature(g.TileIndexFieldName, OGRDS, tileName, points[0], points[1])

    if feature_only:
        return

    s_fh = levelMosaicInfo.getDataSet(
        dec.ulx, dec.uly + height * dec.scaleY, dec.ulx + width * dec.scaleX, dec.uly
    )
    if s_fh is None:
        return

    if g.BandType is None:
        bt = levelMosaicInfo.band_type
    else:
        bt = g.BandType

    geotransform = [dec.ulx, dec.scaleX, 0, dec.uly, 0, dec.scaleY]

    bands = levelMosaicInfo.bands

    if g.MemDriver is None:
        t_fh = g.Driver.Create(temp_tilename, width, height, bands, bt, g.CreateOptions)
    else:
        t_fh = g.MemDriver.Create("", width, height, bands, bt)

    if t_fh is None:
        print("Creation failed, terminating gdal_tile.")
        return 1

    t_fh.SetGeoTransform(geotransform)
    t_fh.SetProjection(levelMosaicInfo.projection)
    for band in range(1, bands + 1):
        t_band = t_fh.GetRasterBand(band)
        if levelMosaicInfo.ct is not None:
            t_band.SetRasterColorTable(levelMosaicInfo.ct)
        t_band.SetRasterColorInterpretation(levelMosaicInfo.ci[band - 1])

    if levelMosaicInfo.nodata is not None:
        for band in range(1, bands + 1):
            t_band = t_fh.GetRasterBand(band)
            t_band.Fill(levelMosaicInfo.nodata)
            t_band.SetNoDataValue(levelMosaicInfo.nodata)

    res = gdal.ReprojectImage(s_fh, t_fh, None, None, g.ResamplingMethod)
    if res != 0:
        print("Reprojection failed for %s, error %d" % (temp_tilename, res))
        return 1

    levelMosaicInfo.closeDataSet(s_fh)

    if g.MemDriver is None:
        t_fh.FlushCache()
    else:
        tt_fh = g.Driver.CreateCopy(temp_tilename, t_fh, 0, g.CreateOptions)
        tt_fh.FlushCache()
        tt_fh = None

    t_fh = None

    if os.path.exists(tileName):
        os.remove(tileName)
    shutil.move(temp_tilename, tileName)

    if g.Verbose:
        print(
            tileName
            + " : "
            + str(offsetX)
            + "|"
            + str(offsetY)
            + "-->"
            + str(width)
            + "-"
            + str(height)
        )


def createTile(
    g, minfo, offsetX, offsetY, width, height, tilename, OGRDS, feature_only
):
    """

    Create tile

    """
    temp_tilename = tilename + ".tmp"

    if g.BandType is None:
        bt = minfo.band_type
    else:
        bt = g.BandType

    dec = AffineTransformDecorator(
        [minfo.ulx, minfo.scaleX, 0, minfo.uly, 0, minfo.scaleY]
    )

    geotransform = [
        dec.ulx + offsetX * dec.scaleX,
        dec.scaleX,
        0,
        dec.uly + offsetY * dec.scaleY,
        0,
        dec.scaleY,
    ]

    if OGRDS is not None:
        dec2 = AffineTransformDecorator(geotransform)
        points = dec2.pointsFor(width, height)
        addFeature(g.TileIndexFieldName, OGRDS, tilename, points[0], points[1])

    if feature_only:
        return

    s_fh = minfo.getDataSet(
        dec.ulx + offsetX * dec.scaleX,
        dec.uly + offsetY * dec.scaleY + height * dec.scaleY,
        dec.ulx + offsetX * dec.scaleX + width * dec.scaleX,
        dec.uly + offsetY * dec.scaleY,
    )
    if s_fh is None:
        return

    bands = minfo.bands

    if g.MemDriver is None:
        t_fh = g.Driver.Create(temp_tilename, width, height, bands, bt, g.CreateOptions)
    else:
        t_fh = g.MemDriver.Create("", width, height, bands, bt)

    if t_fh is None:
        print("Creation failed, terminating gdal_tile.")
        return 1

    t_fh.SetGeoTransform(geotransform)
    if g.Source_SRS is not None:
        t_fh.SetProjection(g.Source_SRS.ExportToWkt())

    readX = min(s_fh.RasterXSize, width)
    readY = min(s_fh.RasterYSize, height)
    for band in range(1, bands + 1):
        s_band = s_fh.GetRasterBand(band)
        t_band = t_fh.GetRasterBand(band)
        if minfo.ct is not None:
            t_band.SetRasterColorTable(minfo.ct)
        if minfo.nodata is not None:
            t_band.Fill(minfo.nodata)
            t_band.SetNoDataValue(minfo.nodata)

        data = s_band.ReadRaster(0, 0, readX, readY, readX, readY, t_band.DataType)
        t_band.WriteRaster(0, 0, readX, readY, data, readX, readY, t_band.DataType)

    minfo.closeDataSet(s_fh)

    if g.MemDriver is None:
        t_fh.FlushCache()
    else:
        tt_fh = g.Driver.CreateCopy(temp_tilename, t_fh, 0, g.CreateOptions)
        tt_fh.FlushCache()
        tt_fh = None

    t_fh = None

    if os.path.exists(tilename):
        os.remove(tilename)
    shutil.move(temp_tilename, tilename)

    if g.Verbose:
        print(
            tilename
            + " : "
            + str(offsetX)
            + "|"
            + str(offsetY)
            + "-->"
            + str(width)
            + "-"
            + str(height)
        )


def createTileIndex(Verbose, dsName, fieldName, srs, driverName):
    OGRDriver = ogr.GetDriverByName(driverName)
    if OGRDriver is None:
        print("ESRI Shapefile driver not found")
        return 1

    OGRDataSource = OGRDriver.Open(dsName)
    if OGRDataSource is not None:
        OGRDataSource.Destroy()
        OGRDriver.DeleteDataSource(dsName)
        if Verbose:
            print("truncating index " + dsName)

    OGRDataSource = OGRDriver.CreateDataSource(dsName)
    if OGRDataSource is None:
        print("Could not open datasource " + dsName)
        return 1

    OGRLayer = OGRDataSource.CreateLayer("index", srs, ogr.wkbPolygon)
    if OGRLayer is None:
        print("Could not create Layer")
        return 1

    OGRFieldDefn = ogr.FieldDefn(fieldName, ogr.OFTString)
    if OGRFieldDefn is None:
        print("Could not create FieldDefn for " + fieldName)
        return 1

    OGRFieldDefn.SetWidth(256)
    if OGRLayer.CreateField(OGRFieldDefn) != 0:
        print("Could not create Field for " + fieldName)
        return 1

    return OGRDataSource


def addFeature(TileIndexFieldName, OGRDataSource, location, xlist, ylist):
    OGRLayer = OGRDataSource.GetLayer()
    OGRFeature = ogr.Feature(OGRLayer.GetLayerDefn())
    if OGRFeature is None:
        print("Could not create Feature")
        return 1

    OGRFeature.SetField(TileIndexFieldName, location)
    wkt = "POLYGON ((%f %f,%f %f,%f %f,%f %f,%f %f ))" % (
        xlist[0],
        ylist[0],
        xlist[1],
        ylist[1],
        xlist[2],
        ylist[2],
        xlist[3],
        ylist[3],
        xlist[0],
        ylist[0],
    )
    OGRGeometry = ogr.CreateGeometryFromWkt(wkt, OGRLayer.GetSpatialRef())
    if OGRGeometry is None:
        print("Could not create Geometry")
        return 1

    OGRFeature.SetGeometryDirectly(OGRGeometry)

    OGRLayer.CreateFeature(OGRFeature)
    OGRFeature.Destroy()


def closeTileIndex(OGRDataSource):
    OGRDataSource.Destroy()


def buildPyramid(g, minfo, createdTileIndexDS, tileWidth, tileHeight, overlap):
    inputDS = createdTileIndexDS
    for level in range(1, g.Levels + 1):
        g.LastRowIndx = -1
        levelMosaicInfo = mosaic_info(minfo.filename, inputDS)
        levelOutputTileInfo = tile_info(
            int(levelMosaicInfo.xsize / 2),
            int(levelMosaicInfo.ysize / 2),
            tileWidth,
            tileHeight,
            overlap,
        )
        inputDS = buildPyramidLevel(g, levelMosaicInfo, levelOutputTileInfo, level)


def buildPyramidLevel(g, levelMosaicInfo, levelOutputTileInfo, level):
    yRange = list(range(1, levelOutputTileInfo.countTilesY + 1))
    xRange = list(range(1, levelOutputTileInfo.countTilesX + 1))

    OGRDS = createTileIndex(
        g.Verbose,
        "TileResult_" + str(level),
        g.TileIndexFieldName,
        g.Source_SRS,
        g.TileIndexDriverTyp,
    )

    for yIndex in yRange:
        for xIndex in xRange:
            offsetY = (yIndex - 1) * (
                levelOutputTileInfo.tileHeight - levelOutputTileInfo.overlap
            )
            offsetX = (xIndex - 1) * (
                levelOutputTileInfo.tileWidth - levelOutputTileInfo.overlap
            )
            height = levelOutputTileInfo.tileHeight
            width = levelOutputTileInfo.tileWidth

            if offsetX + width > levelOutputTileInfo.width:
                width = levelOutputTileInfo.width - offsetX
            if offsetY + height > levelOutputTileInfo.height:
                height = levelOutputTileInfo.height - offsetY

            tilename = getTileName(
                g, levelMosaicInfo, levelOutputTileInfo, xIndex, yIndex, level
            )

            feature_only = g.Resume and os.path.exists(tilename)
            createPyramidTile(
                g,
                levelMosaicInfo,
                offsetX,
                offsetY,
                width,
                height,
                tilename,
                OGRDS,
                feature_only,
            )

    if g.TileIndexName is not None:
        shapeName = getTargetDir(g, level) + g.TileIndexName
        copyTileIndexToDisk(g, OGRDS, shapeName)

    if g.CsvFileName is not None:
        csvName = getTargetDir(g, level) + g.CsvFileName
        copyTileIndexToCSV(g, OGRDS, csvName)

    return OGRDS


def getTileName(g, minfo, ti, xIndex, yIndex, level=-1):
    """
    creates the tile file name
    """
    maxim = ti.countTilesX
    if ti.countTilesY > maxim:
        maxim = ti.countTilesY
    countDigits = len(str(maxim))
    parts = os.path.splitext(os.path.basename(minfo.filename))
    if parts[0][0] == "@":  # remove possible leading "@"
        parts = (parts[0][1 : len(parts[0])], parts[1])

    yIndex_str = ("%0" + str(countDigits) + "i") % (yIndex,)
    xIndex_str = ("%0" + str(countDigits) + "i") % (xIndex,)

    if g.UseDirForEachRow:
        frmt = (
            getTargetDir(g, level)
            + str(yIndex)
            + os.sep
            + parts[0]
            + "_"
            + yIndex_str
            + "_"
            + xIndex_str
        )
        # See if there was a switch in the row, if so then create new dir for row.
        if g.LastRowIndx < yIndex:
            g.LastRowIndx = yIndex
            if not os.path.exists(getTargetDir(g, level) + str(yIndex)):
                os.mkdir(getTargetDir(g, level) + str(yIndex))
    else:
        frmt = getTargetDir(g, level) + parts[0] + "_" + yIndex_str + "_" + xIndex_str
    # Check for the extension that should be used.
    if g.Extension is None:
        frmt += parts[1]
    else:
        frmt += "." + g.Extension
    return frmt


def UsageFormat():
    print("Valid formats:")
    count = gdal.GetDriverCount()
    for index in range(count):
        driver = gdal.GetDriver(index)
        print(driver.ShortName)
    return 1


# =============================================================================


def Usage():
    print("Usage: gdal_retile.py ")
    print("        [-v] [-q] [-co NAME=VALUE]* [-of out_format]")
    print("        [-ps pixelWidth pixelHeight]")
    print("        [-overlap val_in_pixel]")
    print("        [-ot  {Byte/Int16/UInt16/UInt32/Int32/Float32/Float64/")
    print("               CInt16/CInt32/CFloat32/CFloat64}]")
    print("        [ -tileIndex tileIndexName [-tileIndexField fieldName]]")
    print("        [ -csv fileName [-csvDelim delimiter]]")
    print("        [-s_srs srs_def]  [-pyramidOnly] -levels numberoflevels")
    print("        [-r {near/bilinear/cubic/cubicspline/lanczos}]")
    print("        [-useDirForEachRow] [-resume]")
    print("        -targetDir TileDirectory input_files")
    return 2


def main(args=None, g=None):
    if g is None:
        g = RetileGlobals()

    if args is None:
        args = sys.argv
    argv = gdal.GeneralCmdLineProcessor(args)
    if argv is None:
        return 1

    # Parse command line arguments.
    i = 1
    while i < len(argv):
        arg = argv[i]

        if arg == "-of" or arg == "-f":
            i += 1
            g.Format = argv[i]
        elif arg == "-ot":
            i += 1
            g.BandType = gdal.GetDataTypeByName(argv[i])
            if g.BandType == gdal.GDT_Unknown:
                print("Unknown GDAL data type: %s" % argv[i])
                return 1
        elif arg == "-co":
            i += 1
            g.CreateOptions.append(argv[i])

        elif arg == "-v":
            g.Verbose = True
        elif arg == "-q":
            g.Quiet = True

        elif arg == "-targetDir":
            i += 1
            g.TargetDir = argv[i]

            if not os.path.exists(g.TargetDir):
                print("TargetDir " + g.TargetDir + " does not exist")
                return 1
            if g.TargetDir[len(g.TargetDir) - 1 :] != os.sep:
                g.TargetDir = g.TargetDir + os.sep

        elif arg == "-ps":
            i += 1
            g.TileWidth = int(argv[i])
            i += 1
            g.TileHeight = int(argv[i])

        elif arg == "-overlap":
            i += 1
            g.Overlap = int(argv[i])

        elif arg == "-r":
            i += 1
            ResamplingMethodString = argv[i]
            if ResamplingMethodString == "near":
                g.ResamplingMethod = gdal.GRA_NearestNeighbour
            elif ResamplingMethodString == "bilinear":
                g.ResamplingMethod = gdal.GRA_Bilinear
            elif ResamplingMethodString == "cubic":
                g.ResamplingMethod = gdal.GRA_Cubic
            elif ResamplingMethodString == "cubicspline":
                g.ResamplingMethod = gdal.GRA_CubicSpline
            elif ResamplingMethodString == "lanczos":
                g.ResamplingMethod = gdal.GRA_Lanczos
            else:
                print("Unknown resampling method: %s" % ResamplingMethodString)
                return 1
        elif arg == "-levels":
            i += 1
            g.Levels = int(argv[i])
            if g.Levels < 1:
                print("Invalid number of levels : %d" % g.Levels)
                return 1
        elif arg == "-s_srs":
            i += 1
            g.Source_SRS = osr.SpatialReference()
            if g.Source_SRS.SetFromUserInput(argv[i]) != 0:
                print("invalid -s_srs: " + argv[i])
                return 1

        elif arg == "-pyramidOnly":
            g.PyramidOnly = True
        elif arg == "-tileIndex":
            i += 1
            g.TileIndexName = argv[i]
            parts = os.path.splitext(g.TileIndexName)
            if not parts[1]:
                g.TileIndexName += ".shp"

        elif arg == "-tileIndexField":
            i += 1
            g.TileIndexFieldName = argv[i]
        elif arg == "-csv":
            i += 1
            g.CsvFileName = argv[i]
            parts = os.path.splitext(g.CsvFileName)
            if not parts[1]:
                g.CsvFileName += ".csv"
        elif arg == "-csvDelim":
            i += 1
            g.CsvDelimiter = argv[i]
        elif arg == "-useDirForEachRow":
            g.UseDirForEachRow = True
        elif arg == "-resume":
            g.Resume = True
        elif arg[:1] == "-":
            print("Unrecognized command option: %s" % arg)
            return Usage()

        else:
            g.Names.append(arg)
        i += 1

    if not g.Names:
        print("No input files selected.")
        return Usage()

    if g.TileWidth == 0 or g.TileHeight == 0:
        print("Invalid tile dimension %d,%d" % (g.TileWidth, g.TileHeight))
        return 1
    if g.TileWidth - g.Overlap <= 0 or g.TileHeight - g.Overlap <= 0:
        print("Overlap too big w.r.t tile height/width")
        return 1

    if g.TargetDir is None:
        print("Missing Directory for Tiles -targetDir")
        return Usage()

    # create level 0 directory if needed
    if g.UseDirForEachRow and not g.PyramidOnly:
        leveldir = g.TargetDir + str(0) + os.sep
        if not os.path.exists(leveldir):
            os.mkdir(leveldir)

    if g.Levels > 0:  # prepare Dirs for pyramid
        startIndx = 1
        for levelIndx in range(startIndx, g.Levels + 1):
            leveldir = g.TargetDir + str(levelIndx) + os.sep
            if os.path.exists(leveldir):
                continue
            os.mkdir(leveldir)
            if not os.path.exists(leveldir):
                print("Cannot create level dir: %s" % leveldir)
                return 1
            if g.Verbose:
                print("Created level dir: %s" % leveldir)

    g.Driver = gdal.GetDriverByName(g.Format)
    if g.Driver is None:
        print("Format driver %s not found, pick a supported driver." % g.Format)
        return UsageFormat()

    DriverMD = g.Driver.GetMetadata()
    g.Extension = DriverMD.get(gdal.DMD_EXTENSION)
    if "DCAP_CREATE" not in DriverMD:
        g.MemDriver = gdal.GetDriverByName("MEM")

    tileIndexDS = getTileIndexFromFiles(g)
    if tileIndexDS is None:
        print("Error building tile index")
        return 1
    minfo = mosaic_info(g.Names[0], tileIndexDS)
    ti = tile_info(minfo.xsize, minfo.ysize, g.TileWidth, g.TileHeight, g.Overlap)

    if g.Source_SRS is None and minfo.projection:
        g.Source_SRS = osr.SpatialReference()
        if g.Source_SRS.SetFromUserInput(minfo.projection) != 0:
            print("invalid projection  " + minfo.projection)
            return 1

    if g.Verbose:
        minfo.report()
        ti.report()

    if not g.PyramidOnly:
        dsCreatedTileIndex = tileImage(g, minfo, ti)
        tileIndexDS.Destroy()
    else:
        dsCreatedTileIndex = tileIndexDS

    if g.Levels > 0:
        buildPyramid(g, minfo, dsCreatedTileIndex, g.TileWidth, g.TileHeight, g.Overlap)

    if g.Verbose:
        print("FINISHED")
    return 0


class RetileGlobals:
    __slots__ = [
        "Verbose",
        "Quiet",
        "CreateOptions",
        "Names",
        "TileWidth",
        "TileHeight",
        "Overlap",
        "Format",
        "BandType",
        "Driver",
        "Extension",
        "MemDriver",
        "TileIndexFieldName",
        "TileIndexName",
        "TileIndexDriverTyp",
        "CsvDelimiter",
        "CsvFileName",
        "Source_SRS",
        "TargetDir",
        "ResamplingMethod",
        "Levels",
        "PyramidOnly",
        "LastRowIndx",
        "UseDirForEachRow",
        "Resume",
    ]

    def __init__(self):
        """Only used for unit tests"""
        self.Verbose = False
        self.Quiet = False
        self.CreateOptions = []
        self.Names = []
        self.TileWidth = 256
        self.TileHeight = 256
        self.Overlap = 0
        self.Format = "GTiff"
        self.BandType = None
        self.Driver = None
        self.Extension = None
        self.MemDriver = None
        self.TileIndexFieldName = "location"
        self.TileIndexName = None
        self.TileIndexDriverTyp = "Memory"
        self.CsvDelimiter = ";"
        self.CsvFileName = None

        self.Source_SRS = None
        self.TargetDir = None
        self.ResamplingMethod = gdal.GRA_NearestNeighbour
        self.Levels = 0
        self.PyramidOnly = False
        self.LastRowIndx = -1
        self.UseDirForEachRow = False
        self.Resume = False


if __name__ == "__main__":
    sys.exit(main(sys.argv))
